/*
	pppcharge.c - program to start and stop charging of internet
	stuff
	Copyright 1999 Jonathan McDowell for Project Purple

	12/04/1999 - Started.
*/

#define VERSION "0.0.2"

#include <stdio.h>
#include <time.h>
#include <string.h>
#include <unistd.h>
#include <getopt.h>

#ifdef DEBUG
#define CHARGINGFILE "charging.inf"
#define TMPCHARGINGFILE "charging.inf.tmp"
#else
#define CHARGINGFILE "/etc/ppp/charging.inf"
#define TMPCHARGINGFILE "/etc/ppp/charging.inf.tmp"
#endif

struct charginginf {
	char name[9];
	unsigned char action; /* 0 = dialup, 1 = hangup */
	time_t time;
};

static int html=0, verbose=0;

void calcharge(const char *name)
{
	FILE *chgfile;
	struct charginginf chgdata;
	struct tm *timeinfo;
	time_t totaltime;
	time_t lasttime;
	unsigned long totalcalls;
	unsigned long totalcharge,tmpcharge;
	int incharge;

	totaltime=lasttime=totalcalls=totalcharge=0;

	if ((chgfile=fopen(CHARGINGFILE,"r"))==NULL) {
		printf("calcharge(): Couldn't open %s\n", CHARGINGFILE);
		exit(1);
	}

	fread(&chgdata, 1, sizeof(chgdata), chgfile);

	while (!feof(chgfile)) {
		if (name==NULL || strcmp(chgdata.name, name)==0) {
			if (chgdata.action==0) {
				lasttime=chgdata.time;
				++totalcalls;
				incharge=1;
			} else if (chgdata.action==1 && incharge) {
				timeinfo=localtime(&lasttime);

				/* Make tmpcharge equal minutes online */
				tmpcharge = ((chgdata.time-lasttime)+59)/60;

				/* FIXME: This should all be controlled from
				   a config file. */
				if (timeinfo->tm_wday==0 || timeinfo->tm_wday==6)
					/* Weekend. 1p/min */
					;
				else if (timeinfo->tm_hour<8 || timeinfo->tm_hour>17)
					/* 6pm-8am. 1.5p/min */
					tmpcharge *= 1.5;
				else
					/* 8am-6pm. 4p/min */
					tmpcharge *= 4;

				/* BT minimum call charge */
				if (tmpcharge<5) tmpcharge=5;

				if (verbose) {
					if (chgdata.time-lasttime < 60)
						printf("Call charge of %ldp for %ld secs.\n", tmpcharge, chgdata.time-lasttime);
					else
						printf("Call charge of %ldp for %ld mins, %ld secs.\n", tmpcharge, (chgdata.time-lasttime)/60, (chgdata.time-lasttime) % 60);
				}

				totalcharge+=tmpcharge;

				totaltime+=(chgdata.time-lasttime);
				incharge=0;
			}
		}

		fread(&chgdata, 1, sizeof(chgdata), chgfile);
	}

	fclose(chgfile);

	if (name==NULL)
		printf("Global call details:%s\n", html ? "<BR>" : "");
	else
		printf("Call details for %s:%s\n", name, html ? "<BR>" : "");

	if (totaltime<60)
		printf("\tTotal time online: %ld secs%s\n", totaltime, html ? "<BR>" : "");
	else if (totaltime<3600)
		printf("\tTotal time online %ld mins, %ld secs%s\n", totaltime/60, totaltime % 60, html ? "<BR>" : "");
	else
		printf("\tTotal time online %ld hours, %ld mins, %ld secs%s\n", totaltime/3600, ((totaltime % 3600)/60), totaltime % 60, html ? "<BR>" : "");

	printf("\tTotal calls: %ld%s\n", totalcalls, html ? "<BR>" : "");
	printf("\tTotal charge: %ldp%s\n", totalcharge, html ? "<BR>" : "");
}

void curcharge(void)
{
	FILE *chgfile;
	struct charginginf chgdata;
	struct tm *timeinfo;
	time_t totaltime;
	time_t lasttime;
	unsigned long tmpcharge;

	totaltime=lasttime=0;

	if ((chgfile=fopen(CHARGINGFILE,"r"))==NULL) {
		printf("curcharge(): Couldn't open %s\n", CHARGINGFILE);
		exit(1);
	}

	fseek(chgfile, -sizeof(chgdata), SEEK_END);
	fread(&chgdata, 1, sizeof(chgdata), chgfile);

	if (chgdata.action==0) {
		timeinfo=localtime(&chgdata.time);

		/* Make tmpcharge equal minutes online */
		tmpcharge = ((time(NULL)-chgdata.time)+59)/60;

		if (timeinfo->tm_wday==0 || timeinfo->tm_wday==6)
			/* Weekend. 1p/min */
			;
		else if (timeinfo->tm_hour<8 || timeinfo->tm_hour>17)
			/* 6pm-8am. 1.5p/min */
			tmpcharge *= 1.5;
		else
			/* 8am-6pm. 4p/min */
			tmpcharge *= 4;

		/* BT minimum call charge */
		if (tmpcharge<5) tmpcharge=5;

		printf("Current call has lasted %ld secs, cost so far %ldp\n",
			time(NULL)-chgdata.time, tmpcharge);

	} else if (chgdata.action==1) {
		printf("Not currently in a call.\n");
	}

	fclose(chgfile);
}

void fixcharging()
{
	FILE *oldchgfile, *newchgfile;
	struct charginginf chgdata, oldchgdata;
	int incharge;

	if ((oldchgfile=fopen(CHARGINGFILE,"r"))==NULL) {
		printf("fixcharging(): Couldn't open %s\n", CHARGINGFILE);
		exit(1);
	}

	if ((newchgfile=fopen(TMPCHARGINGFILE,"w"))==NULL) {
		printf("fixcharging(): Couldn't open %s\n", TMPCHARGINGFILE);
		exit(1);
	}

	fread(&chgdata, 1, sizeof(chgdata), oldchgfile);
	incharge=0;

	while (!feof(oldchgfile)) {
		if (chgdata.action==0 && !incharge)
			incharge=1;
		else if (chgdata.action==0 && incharge) {
			printf("Found new dialup action when previous one was never hung up. Fixing.\n");
			oldchgdata.time=chgdata.time-1;
			oldchgdata.action=1;
			fwrite(&oldchgdata, 1, sizeof(oldchgdata), newchgfile);
		} else if (chgdata.action==1) incharge=0;
	
		fwrite(&chgdata, 1, sizeof(chgdata), newchgfile);
		oldchgdata=chgdata;
		fread(&chgdata, 1, sizeof(chgdata), oldchgfile);
	}

	fclose(newchgfile);
	fclose(oldchgfile);

	rename(TMPCHARGINGFILE, CHARGINGFILE);
}

void listcharging(const char *name)
{
	FILE *chgfile;
	struct charginginf chgdata;

	if ((chgfile=fopen(CHARGINGFILE,"r"))==NULL) {
		printf("listcharging(): Couldn't open %s\n", CHARGINGFILE);
		exit(1);
	}

	if (html) {
		printf("<TABLE>\n");
		printf("<TR><TH>Name</TH><TH>Action</TH><TH>Time</TH></TR>\n");
	} else
		printf("Name     Action  Time\n");

	fread(&chgdata, 1, sizeof(chgdata), chgfile);

	while (!feof(chgfile)) {
		if (name==NULL || strcmp(name, chgdata.name)==0)
			if (html)
		printf("<TR><TD>%s</TD><TD>%s</TD><TD>%ld %s</TD></TR>",chgdata.name, (chgdata.action==0) ?
			"started" : "hungup ", chgdata.time, ctime(&chgdata.time));
			else
		printf("%-8s %s %ld %s",chgdata.name, (chgdata.action==0) ?
			"started" : "hungup ", chgdata.time, ctime(&chgdata.time));

		fread(&chgdata, 1, sizeof(chgdata), chgfile);
	}

	fclose(chgfile);
	
	if (html) printf("</TABLE>\n");
}

void startstop(const char *name, unsigned char action)
{
	FILE *chgfile;
	struct charginginf chgdata;

	if ((chgfile=fopen(CHARGINGFILE,"a"))==NULL) {
		printf("startstop(): Couldn't open %s\n", CHARGINGFILE);
		exit(1);
	}

	strncpy(chgdata.name, name, 8);
	chgdata.name[8]=0;
	chgdata.action=action;
	chgdata.time=time(NULL);

	fwrite(&chgdata, 1, sizeof(chgdata), chgfile);

	fclose(chgfile);
}

void usage()
{
	printf("pppcharge v%s by Jonathan McDowell\n", VERSION);
	printf("usage: pppcharge [OPTION]\n\n");
	printf("  --start         Start charging <user>\n");
	printf("  --stop          Stop charging <user>\n");
	printf("  --bill          Calculate bill for given user, or total bill if no user\n");
	printf("                  given.\n");
	printf("  --current       Calculate the cost of the current call\n");
	printf("  --help          Display this usage information.\n");
	printf("  --list          List only entries for given user, or all entries if no user\n");
	printf("                  given.\n");
	printf("  --user <user>   Specify which user to perform the given action for.\n");
	printf("  --html          Use html formatting codes when outputting data. Note that\n");
	printf("                  this does not output header or footer html.\n");
}

int main(int argc, char **argv)
{
	int option;
	char *user=NULL;
	int function=0;
	
	while (1) {
		int longoption;
		static struct option long_options[]={
			{"bill", 0, NULL, 'b'},
			{"current", 0, NULL, 'c'},
			{"help", 0, NULL, 'h'},
			{"html", 0, &html, 1},
			{"list", 0, NULL, 'l'},
			{"start", 0, NULL, 'x'},
			{"stop", 0, NULL, 'y'},
			{"user", 1, NULL, 'u'},
			{"verbose", 0, &verbose, 1},
			{0,0,0,0}
		};

		option=getopt_long(argc, argv, "bcfhltu:", long_options, &longoption);

		if (option==-1) break;

		switch (option) {
		case 'b':	function+=1; /* bill */
				break;
		case 'c':	function+=2; /* current */
				break;
		case 'f':	function+=4; /* fix */
				break;
		case 'h':	function+=8; /* usage */
				break;
		case 'l':	function+=16; /* list */
				break;
		case 'u':	if (user==NULL)
					user=strdup(optarg);
				else
					printf("Already set user!\n");
				break;
		case 'x':	function+=32;
				break;
		case 'y':	function+=64;
				break;
		}
	}

#ifdef DEBUG
	if (html) {
		puts("<!-- Development version -->");
	} else {
		puts("***DEVELOPMENT VERSION***");
	}
#endif

	switch (function) {
	case 0:		printf("No function specified!\n");
			break;
	case 1:		calcharge(user);
			break;
	case 2:		curcharge();
			break;
	case 4:		fixcharging();
			break;
	case 8:		usage();
			break;
	case 16:	listcharging(user);
			break;
	case 32:	if (user==NULL || strlen(user)==0) {
				puts("Must set user to start charging.");
				exit(1);
			}
			startstop(user, 0);	/* start */
			break;
	case 64:	if (user==NULL || strlen(user)==0) {
				puts("Must set user to stop charging.");
			}
			startstop(user, 1);	/* stop */
			break;
	default:	printf("Can't specify more than one function!\n");
			break;
	}

	return 0;
}
